﻿using UnityEngine;
using System.Collections;

public class Game : MonoBehaviour {

	public Transform theCamera;
	public Player thePlayer;
	public Goal theGoal;
	public Transform cameraFrontPosition;
	public Transform cameraSidePosition;

	public string nextLevel;

	private Side currentSide;

	private Matrix4x4 ortho, perspective;
	public float fov = 60f, near = .3f, far = 1000f, orthographicSize = 5f;
	private float aspect;
	private MatrixBlender blender;

	private float lastAspect;

	// Use this for initialization
	void Start () {
		StartCoroutine(InitializeSide(Side.Front));
	}
	
	// Update is called once per frame
	void Update () {
		
		if(thePlayer.inputBlocked) {
			return;
		}
			
		aspect = (float) Screen.width / (float) Screen.height;
		if(lastAspect != aspect) {
			ortho = Matrix4x4.Ortho(-orthographicSize * aspect, orthographicSize * aspect, -orthographicSize, orthographicSize, near, far);
			perspective = Matrix4x4.Perspective(fov, aspect, near, far);
			theCamera.camera.projectionMatrix = ortho;
			blender = (MatrixBlender) theCamera.GetComponent(typeof(MatrixBlender));
		}

		if(Input.GetKeyDown(KeyCode.Space)) {
			if(currentSide == Side.Side) {
				StartCoroutine(SetSide(Side.Front));
			} else {
				StartCoroutine(SetSide(Side.Side));
			}
		}

		Vector2 goalVector;
		if(currentSide == Side.Front) {
			goalVector = 
				new Vector2(thePlayer.transform.position.x, thePlayer.transform.position.y) - 
				new Vector2(theGoal.transform.position.x, theGoal.transform.position.y);
		} else {
			goalVector = 
				new Vector2(thePlayer.transform.position.z, thePlayer.transform.position.y) - 
				new Vector2(theGoal.transform.position.z, theGoal.transform.position.y);
		}
		if(goalVector.magnitude < 1.0f) {
			StartCoroutine(Goaled());
		}
	}

	IEnumerator Goaled() {
		thePlayer.inputBlocked = true;
		blender.BlendToMatrix(Matrix4x4.Perspective(170, 170, near, far), 2.0f);
		yield return new WaitForSeconds(1.0f);

		if(nextLevel != null && nextLevel.Length != 0) {
			Application.LoadLevel (nextLevel); 
		} else {
			Application.Quit();
		}
	}

	IEnumerator InitializeSide(Side side) {
		yield return null;
		this.currentSide = side;
		GameObject[] walls = GameObject.FindGameObjectsWithTag("Wall");
		foreach(GameObject wall in walls) {
			wall.GetComponent<Wall>().transformMe(this.currentSide);
		}
		thePlayer.SetSide(side);
	}

	IEnumerator SetSide(Side side) {

		thePlayer.inputBlocked = true;
		yield return null;

		this.currentSide = side;

		GameObject[] walls = GameObject.FindGameObjectsWithTag("Wall");
		foreach(GameObject wall in walls) {
			wall.GetComponent<Wall>().resetMe();
		}

		blender.BlendToMatrix(perspective, 0.25f);
		yield return new WaitForSeconds(0.25f);

		float duration = 0.5f;

		if(side == Side.Front) {
			LeanTween.rotate(theCamera.gameObject,  Vector3.zero, duration);
			LeanTween.move(theCamera.gameObject, cameraFrontPosition.transform.position, duration);
		} else {
			LeanTween.rotate(theCamera.gameObject,  new Vector3(0.0f, -90.0f, 0.0f), duration);
			LeanTween.move(theCamera.gameObject, cameraSidePosition.transform.position, duration);
		}


		yield return new WaitForSeconds(0.5f);
		blender.BlendToMatrix(ortho, 0.25f);
		yield return new WaitForSeconds(0.25f);

		foreach(GameObject wall in walls) {
			wall.GetComponent<Wall>().transformMe(side);
		}

		thePlayer.SetSide(side);
		thePlayer.inputBlocked = false;
	}
}
